/* 
 * mode.c -- handles:
 * queueing and flushing mode changes made by the bot
 * channel mode changes and the bot's reaction to them
 * setting and getting the current wanted channel modes
 * dprintf'ized, 12dec1995
 * multi-channel, 6feb1996
 * stopped the bot deopping masters and bots in bitch mode, pteron 23Mar1997
 */
/* 
 * This file is part of the eggdrop source code
 * copyright (c) 1997 Robey Pointer
 * and is distributed according to the GNU general public license.
 * For full details, read the top of 'main.c' or the file called
 * COPYING that was distributed with this code.
 */

/* reversing this mode? */
static int reversing = 0;

#define PLUS    1
#define MINUS   2
#define CHOP    4
#define BAN     8
#define VOICE   16
#define EXEMPT  32
#define INVITE  64

static struct flag_record user = {FR_GLOBAL | FR_CHAN, 0, 0, 0, 0, 0};
static struct flag_record victim = {FR_GLOBAL | FR_CHAN, 0, 0, 0, 0, 0};

static void flush_mode(struct chanset_t *chan, int pri)
{
  char *p, out[512], post[512];
  int i, ok = 0;

  p = out;
  post[0] = 0;
  if (chan->pls[0])
    *p++ = '+';
  for (i = 0; i < strlen(chan->pls); i++)
    *p++ = chan->pls[i];
  if (chan->mns[0])
    *p++ = '-';
  for (i = 0; i < strlen(chan->mns); i++)
    *p++ = chan->mns[i];
  chan->pls[0] = 0;
  chan->mns[0] = 0;
  chan->bytes = 0;
  chan->compat = 0;
  ok = 0;
  /* +k or +l ? */
  if (chan->key[0]) {
    if (!ok) {
      *p++ = '+';
      ok = 1;
    }
    *p++ = 'k';
    strcat(post, chan->key);
    strcat(post, " ");
  }
  if (chan->limit != (-1)) {
    if (!ok) {
      *p++ = '+';
      ok = 1;
    }
    *p++ = 'l';
    sprintf(&post[strlen(post)], "%d ", chan->limit);
  }
  chan->limit = (-1);
  chan->key[0] = 0;
  /* do -b before +b to avoid server ban overlap ignores */
  for (i = 0; i < modesperline; i++)
    if ((chan->cmode[i].type & MINUS) && (chan->cmode[i].type & BAN)) {
      if (ok < 2) {
	*p++ = '-';
	ok = 2;
      }
      *p++ = 'b';
      strcat(post, chan->cmode[i].op);
      strcat(post, " ");
      nfree(chan->cmode[i].op);
      chan->cmode[i].op = NULL;
    }
  ok &= 1;
  for (i = 0; i < modesperline; i++)
    if (chan->cmode[i].type & PLUS) {
      if (!ok) {
	*p++ = '+';
	ok = 1;
      }
      *p++ = ((chan->cmode[i].type & BAN) ? 'b' :
	      ((chan->cmode[i].type & CHOP) ? 'o' : 
	       ((chan->cmode[i].type & EXEMPT) ? 'e' : 
		((chan->cmode[i].type & INVITE) ? 'I' : 'v'))));
      strcat(post, chan->cmode[i].op);
      strcat(post, " ");
      nfree(chan->cmode[i].op);
      chan->cmode[i].op = NULL;
    }
  ok = 0;
  /* -k ? */
  if (chan->rmkey[0]) {
    if (!ok) {
      *p++ = '-';
      ok = 1;
    }
    *p++ = 'k';
    strcat(post, chan->rmkey);
    strcat(post, " ");
  }
  chan->rmkey[0] = 0;
  for (i = 0; i < modesperline; i++)
    if ((chan->cmode[i].type & MINUS) && !(chan->cmode[i].type & BAN)) {
      if (!ok) {
	*p++ = '-';
	ok = 1;
      }
	 *p++ = ((chan->cmode[i].type & CHOP) ? 'o' : 
		 ((chan->cmode[i].type & EXEMPT) ? 'e' : 
		  ((chan->cmode[i].type & INVITE) ? 'I' : 'v')));
      strcat(post, chan->cmode[i].op);
      strcat(post, " ");
      nfree(chan->cmode[i].op);
      chan->cmode[i].op = NULL;
    }
  *p = 0;
  for (i = 0; i < modesperline; i++)
    chan->cmode[i].type = 0;
  if (post[0] && post[strlen(post) - 1] == ' ')
    post[strlen(post) - 1] = 0;
  if (post[0]) {
    strcat(out, " ");
    strcat(out, post);
  }
  if (out[0]) {
    if (pri == QUICK)
      dprintf(DP_MODE, "MODE %s %s\n", chan->name, out);
    else
      dprintf(DP_SERVER, "MODE %s %s\n", chan->name, out);
  }
}

/* queue a channel mode change */
static void real_add_mode(struct chanset_t *chan, char plus, char mode, char *op)
{
  int i, type, ok, l;
  int bans, exempts, invites, modes;
  masklist *m;
  memberlist *mx;
  char s[21];
  
  Context;
  if (!me_op(chan))
    return;			/* no point in queueing the mode */

  if (mode == 'o' || mode == 'v') {
    mx = ismember(chan, op);
    if (!mx)
      return;

    if (plus == '-' && mode == 'o') {
      if (chan_sentdeop(mx) || !chan_hasop(mx))
	return;
      mx->flags |= SENTDEOP;
    }
    if (plus == '+' && mode == 'o') {
      if (chan_sentop(mx) || chan_hasop(mx))
	return;
      mx->flags |= SENTOP;
    }
    if (plus == '-' && mode == 'v') {
      if (chan_sentdevoice(mx) || !chan_hasvoice(mx))
	return;
      mx->flags |= SENTDEVOICE;
    }
    if (plus == '+' && mode == 'v') {
      if (chan_sentvoice(mx) || chan_hasvoice(mx))
	return;
      mx->flags |= SENTVOICE;
    }
  }

  if (chan->compat == 0) {
    if (mode == 'e' || mode == 'I')
      chan->compat = 2;
    else
      chan->compat = 1;
  } else if (mode == 'e' || mode == 'I') {
    if (prevent_mixing && chan->compat == 1)
      flush_mode(chan, NORMAL);
  } else if (prevent_mixing && chan->compat == 2)
    flush_mode(chan, NORMAL);

  if ((mode == 'o') || (mode == 'b') || (mode == 'v') ||
      (mode == 'e') || (mode == 'I')) {
    type = (plus == '+' ? PLUS : MINUS) |
      (mode == 'o' ? CHOP : (mode == 'b' ? BAN : (mode == 'v' ? VOICE : (mode == 'e' ? EXEMPT : INVITE))));
    /* if -b'n a non-existant ban...nuke it */
    if ((plus == '-') && (mode == 'b'))
/* FIXME: some network remove overlapped bans, IrcNet doesnt (poptix/drummer)*/
      /*if (!isbanned(chan, op))*/
      if (!ischanban(chan, op))
	return;
    /* if there are already max_bans bans on the channel, don't try to add 
     * one more */
    Context;
    bans = 0;
    for (m = chan->channel.ban; m && m->mask[0]; m = m->next)
      bans++;
    Context;
    if ((plus == '+') && (mode == 'b'))
      if (bans >= max_bans)
	return;
    /* if there are already max_exempts exemptions on the channel, don't
     * try to add one more */
    Context;
    exempts = 0;
    for (m = chan->channel.exempt; m && m->mask[0]; m = m->next)
      exempts++;
    Context;
    if ((plus == '+') && (mode == 'e'))
      if (exempts >= max_exempts)
	return;
    /* if there are already max_invites invitations on the channel, don't
     * try to add one more */
    Context;
    invites = 0;
    for (m = chan->channel.invite; m && m->mask[0]; m = m->next)
      invites++;
    Context;
    if ((plus == '+') && (mode == 'I'))
      if (invites >= max_invites)
	return;
    /* if there are already max_modes +b/+e/+I modes on the channel, don't 
     * try to add one more */
    modes = bans + exempts + invites;
    if ((modes >= max_modes) && (plus == '+') &&
	((mode == 'b') || (mode == 'e') || (mode == 'I')))
      return;
    /* op-type mode change */
    for (i = 0; i < modesperline; i++)
      if ((chan->cmode[i].type == type) && (chan->cmode[i].op != NULL) &&
	  (!rfc_casecmp(chan->cmode[i].op, op)))
	return;			/* already in there :- duplicate */
    ok = 0;			/* add mode to buffer */
    l = strlen(op) + 1;
    if ((chan->bytes + l) > mode_buf_len)
      flush_mode(chan, NORMAL);
    for (i = 0; i < modesperline; i++)
      if ((chan->cmode[i].type == 0) && (!ok)) {
	chan->cmode[i].type = type;
	chan->cmode[i].op = (char *) channel_malloc(l);
	chan->bytes += l;	/* add 1 for safety */
	strcpy(chan->cmode[i].op, op);
	ok = 1;
      }
    ok = 0;			/* check for full buffer */
    for (i = 0; i < modesperline; i++)
      if (chan->cmode[i].type == 0)
	ok = 1;
    if (!ok)
      flush_mode(chan, NORMAL);	/* full buffer!  flush modes */
    if ((mode == 'b') && (plus == '+') && channel_enforcebans(chan))
      enforce_bans(chan);
/* recheck_channel(chan,0); */
    return;
  }
  /* +k ? store key */
  if ((plus == '+') && (mode == 'k'))
    strcpy(chan->key, op);
  /* -k ? store removed key */
  else if ((plus == '-') && (mode == 'k'))
    strcpy(chan->rmkey, op);
  /* +l ? store limit */
  else if ((plus == '+') && (mode == 'l'))
    chan->limit = atoi(op);
  else {
    /* typical mode changes */
    if (plus == '+')
      strcpy(s, chan->pls);
    else
      strcpy(s, chan->mns);
    if (!strchr(s, mode)) {
      if (plus == '+') {
	chan->pls[strlen(chan->pls) + 1] = 0;
	chan->pls[strlen(chan->pls)] = mode;
      } else {
	chan->mns[strlen(chan->mns) + 1] = 0;
	chan->mns[strlen(chan->mns)] = mode;
      }
    }
  }
}

/**********************************************************************/
/* horrible code to parse mode changes */
/* no, it's not horrible, it just looks that way */

static void got_key(struct chanset_t *chan, char *nick, char *from, char *key)
{
    int remove=0;
    
    if (!nick[0] || (chan->mode_mns_prot & CHANKEY)) remove=1;
    set_key(chan, key);
    if (remove && ((rand() % 100) < ameno_op_prc)) {
	if (strlen(key) != 0) {
	    add_mode(chan, '-', 'k', key);
	} else {
	    add_mode(chan, '-', 'k', "");
	}
    }
}

static void got_op(struct chanset_t *chan, char *nick, char *from, char *who, struct userrec *opu, struct flag_record *opper)
{
  memberlist *t;
  memberlist *m;
  char s[UHOSTLEN];
  struct userrec *u;
  int check_chan = 0;

  m = ismember(chan, who);
  t = ismember(chan, nick);
  if (!m) {
    ameno_addkick(chan->name, who);
    return;
  }
  /* Did *I* just get opped? */
  if (!me_op(chan) && match_my_nick(who)) check_chan=1;

  if (!m->user) {
    simple_sprintf(s, "%s!%s", m->nick, m->userhost);
    u = get_user_by_host(s);
  } else u = m->user;

  get_user_flagrec(u, &victim, chan->name);
  /* flags need to be set correctly right from the beginning now, so that
   * add_mode() doesn't get irritated  (Fabian) */
  m->flags |= CHANOP;
  check_tcl_mode(nick, from, u, chan->name, "+o", who);
  /* added new meaning of WASOP:
   * in mode binds it means: was he op before get (de)opped
   * (stupid IrcNet allows opped users to be opped again and
   *  opless users to be deopped)
   * script now can use [wasop nick chan] proc to check
   * if user was op or wasnt  (drummer) */
  m->flags &= ~SENTOP;

  if (me_op(chan) && !match_my_nick(who) && nick[0]) {
    if (((channel_masskick(chan) && !glob_op(victim) && !chan_op(victim)) || (channel_bitch(chan) && !glob_op(victim) && !chan_op(victim) && !glob_master(*opper) && !chan_master(*opper))) && !(m->flags & WASOP)) {
       m->flags |= SENTKICK;
       t->flags |= SENTKICK;
       ameno_addkick(chan->name, who);
       ameno_addkick(chan->name, nick);
    } else if ((chan_deop(victim) || (glob_deop(victim) && !chan_op(victim)))) ameno_addkick(chan->name, who);
  }
  if (!nick[0] && me_op(chan) && !match_my_nick(who)) {
    if (channel_bitch(chan) && !glob_op(victim) && !chan_op(victim)) {
       m->flags|=SENTKICK;
       ameno_addkick(chan->name, who);
    }
    if (chan_deop(victim) || (glob_deop(victim) && !chan_op(victim))) {
      m->flags|=SENTKICK;
      ameno_addkick(chan->name, who);
    }
  }
  m->flags|=WASOP;
  if (check_chan) recheck_channel(chan, 1);
  ameno_delop(chan->name, who);
}

static void got_deop(struct chanset_t *chan, char *nick, char *from, char *who, struct userrec *opu)
{
  memberlist *m;
  char s[UHOSTLEN], s1[UHOSTLEN];
  struct userrec *u;
  int had_op;

  m = ismember(chan, who);
  if (!m) return;
  simple_sprintf(s, "%s!%s", m->nick, m->userhost);
  simple_sprintf(s1, "%s!%s", nick, from);
  u = get_user_by_host(s);
  get_user_flagrec(u, &victim, chan->name);

  had_op = chan_hasop(m);
  /* flags need to be set correctly right from the beginning now, so that
   * add_mode() doesn't get irritated  (Fabian) */
  m->flags &= ~(CHANOP | SENTDEOP | FAKEOP);
  check_tcl_mode(nick, from, u, chan->name, "-o", who);
  /* check comments in got_op()  (drummer) */
  m->flags &= ~WASOP;

  if (me_op(chan)) {
    if (channel_massop(chan) && glob_op(victim) && glob_bot(victim) && !match_my_nick(nick) && rfc_casecmp(who, nick) && had_op && !match_my_nick(who)) {
      ameno_addop(chan->name, who);
    }
  }
  /* was the bot deopped? */
  if (match_my_nick(who)) {
    /* cancel any pending kicks and modes */
    memberlist *m2 = chan->channel.member;

    while (m2 && m2->nick[0]) {
	m2->flags &= ~(SENTKICK | SENTDEOP | SENTOP | SENTVOICE | SENTDEVOICE);
      m2 = m2->next;
    }
    if (chan->need_op[0]) do_tcl("need-op", chan->need_op);
    if (strcasecmp(chan->name, ameno_kick_chan) == 0) ameno_kick_clear();
    if (strcasecmp(chan->name, ameno_op_chan) == 0) ameno_op_clear();
  }
  if (nick[0]) maybe_revenge(chan, s1, s, REVENGE_DEOP);
}


static void got_ban(struct chanset_t *chan, char *nick, char *from, char *who)
{
    char me[UHOSTLEN], s[UHOSTLEN], s1[UHOSTLEN];
    int check=1;
    int remove=0;
    int kick=0;
    memberlist *m;
    struct userrec *u;

    Context;
    simple_sprintf(me, "%s!%s", botname, botuserhost);
    simple_sprintf(s, "%s!%s", nick, from);
    newban(chan, who, s);
    if (wild_match(who, me) && me_op(chan)) {
	remove=1;
	check=0;
	if (!match_my_nick(nick) && !glob_master(user) && !glob_bot(user) && !chan_master(user)) kick=1;
    } else if (!match_my_nick(nick)) {	/* it's not my ban */
	if (channel_nouserbans(chan) && nick[0] && !glob_bot(user) && !glob_master(user) && !chan_master(user) && (!u_equals_mask(global_bans, who) && !u_equals_mask(chan->bans, who))) {
	    kick=1;
	    remove=1;
	    check=0;
	} else if (!nick[0] && (!u_equals_mask(global_bans, who) && !u_equals_mask(chan->bans, who))) {
	    remove=1;
	    check=0;
	}
	u=get_user_by_host(who);
	if (u) {
	    get_user_flagrec(u, &victim, chan->name);
	    if (glob_master(victim) || chan_master(victim) || glob_bot(victim)) {
	        remove=1;
		check=0;
	        if (!glob_master(user) && !glob_bot(user) && !chan_master(user)) kick=1;
	    } else if ((glob_op(victim) && !chan_deop(victim)) || chan_op(victim)) {
	        if (!glob_master(user) && !glob_bot(user) && !chan_master(user)) {
		    remove=1;
		    if (!(glob_op(user) && !chan_deop(user)) && !chan_op(user)) kick=1;
		}
	    }
        }
        if (check) {
	    m=chan->channel.member;
    	    while (m && m->nick[0]) {
		sprintf(s1, "%s!%s", m->nick, m->userhost);
		if (wild_match(who, s1)) {
	    	    u=get_user_by_host(s1);
	    	    if (u) {
			get_user_flagrec(u, &victim, chan->name);
			if (glob_master(victim) || chan_master(victim) || glob_bot(victim)) {
			    remove=1;
			    if (!glob_master(user) && !glob_bot(user) && !chan_master(user)) kick=1;
			} else if ((glob_op(victim) && !chan_deop(victim)) || chan_op(victim)) {
			    if (!glob_master(user) && !glob_bot(user) && !chan_master(user)) {
				remove=1;
				if (!(glob_op(user) && !chan_deop(user)) && !chan_op(user)) kick=1;
			    }
			} else if (!u_equals_mask(global_exempts, who) && !u_equals_mask(chan->exempts, who)) ameno_addkick(chan->name, m->nick);
		    } else if (!u_equals_mask(global_exempts, who) && !u_equals_mask(chan->exempts, who)) ameno_addkick(chan->name, m->nick);
		}
		m = m->next;
	    }
	}
    }
    if (remove && ((rand() % 100) < ameno_op_prc)) add_mode(chan, '-', 'b', who);
    if (kick) ameno_addkick(chan->name, nick);
    refresh_exempt(chan,who);
}

static void got_unban(struct chanset_t *chan, char *nick, char *from, char *who, struct userrec *u)
{
    masklist *b, *old;

    Context;
    b=chan->channel.ban;
    old=NULL;
    while (b->mask[0] && rfc_casecmp(b->mask, who)) {
	old=b;
	b=b->next;
    }
    if (b->mask[0]) {
	if (old) old->next = b->next; else chan->channel.ban = b->next;
	nfree(b->mask);
	nfree(b->who);
	nfree(b);
    }
    if (u_sticky_mask(chan->bans, who) || u_sticky_mask(global_bans, who)) {
	if ((rand() % 100) < ameno_op_prc) add_mode(chan, '+', 'b', who);
        if (u) {
	    get_user_flagrec(u, &victim, chan->name);
	    if (!glob_master(victim) && !chan_master(victim) && !glob_bot(victim) && !match_my_nick(nick)) ameno_addkick(chan->name, nick);
	} else ameno_addkick(chan->name, nick);
    }	
}

static void got_exempt(struct chanset_t *chan, char *nick, char *from, char *who)
{
    char me[UHOSTLEN], s[UHOSTLEN];
    int remove=0;
    int kick=0;

    Context;
    simple_sprintf(me, "%s!%s", botname, botuserhost);
    simple_sprintf(s, "%s!%s", nick, from);
    newexempt(chan, who, s);
    if (!match_my_nick(nick)) {	/* it's not my exempt */
	if (channel_nouserexempts(chan) && nick[0] && !glob_bot(user) && !glob_master(user) && !chan_master(user) && (!u_equals_mask(global_exempts, who) && !u_equals_mask(chan->exempts, who))) {
	    kick=1;
	    remove=1;
	} else if (!nick[0] && (!u_equals_mask(global_exempts, who) && !u_equals_mask(chan->exempts, who))) {
	    remove=1;
	}
    }
    if (remove && ((rand() % 100) < ameno_op_prc)) add_mode(chan, '-', 'e', who);
    if (kick) ameno_addkick(chan->name, nick);
}

static void got_unexempt(struct chanset_t *chan, char *nick, char *from, char *who, struct userrec *u)
{
    masklist *e, *old;

    Context;
    e=chan->channel.exempt;
    old=NULL;
    while (e && e->mask[0] && rfc_casecmp(e->mask, who)) {
	old=e;
	e=e->next;
    }
    if (e && e->mask[0]) {
	if (old) old->next = e->next; else chan->channel.exempt = e->next;
	nfree(e->mask);
	nfree(e->who);
	nfree(e);
    }
    if (u_sticky_mask(chan->exempts, who) || u_sticky_mask(global_exempts, who)) {
	if ((rand() % 100) < ameno_op_prc) add_mode(chan, '+', 'e', who);
        if (u) {
	    get_user_flagrec(u, &victim, chan->name);
	    if (!glob_master(victim) && !chan_master(victim) && !glob_bot(victim) && !match_my_nick(nick)) ameno_addkick(chan->name, nick);
	} else ameno_addkick(chan->name, nick);
    } 
}

static void got_invite(struct chanset_t *chan, char *nick, char *from, char *who)
{
    char me[UHOSTLEN], s[UHOSTLEN];
    int remove=0;
    int kick=0;

    Context;
    simple_sprintf(me, "%s!%s", botname, botuserhost);
    simple_sprintf(s, "%s!%s", nick, from);
    newinvite(chan, who, s);
    if (!match_my_nick(nick)) {	/* it's not my exempt */
	if (channel_nouserinvites(chan) && nick[0] && !glob_bot(user) && !glob_master(user) && !chan_master(user) && (!u_equals_mask(global_exempts, who) && !u_equals_mask(chan->exempts, who))) {
	    kick=1;
	    remove=1;
	} else if (!nick[0] && (!u_equals_mask(global_invites, who) && !u_equals_mask(chan->invites, who))) {
	    remove=1;
	}
    }
    if (remove && ((rand() % 100) < ameno_op_prc)) add_mode(chan, '-', 'I', who);
    if (kick) ameno_addkick(chan->name, nick);
}

static void got_uninvite(struct chanset_t *chan, char *nick, char *from,
			 char *who, struct userrec *u)
{
    masklist *inv, *old;

    inv=chan->channel.invite;
    old=NULL;
    while (inv->mask[0] && rfc_casecmp(inv->mask, who)) {
	old=inv;
	inv=inv->next;
    }
    if (inv->mask[0]) {
	if (old) old->next = inv->next; else chan->channel.invite = inv->next;
	nfree(inv->mask);
        nfree(inv->who);
        nfree(inv);
    }
    if (u_sticky_mask(chan->invites, who) || u_sticky_mask(global_invites, who)) {
	if ((rand() % 100) < ameno_op_prc) add_mode(chan, '+', 'I', who);
        if (u) {
	    get_user_flagrec(u, &victim, chan->name);
	    if (!glob_master(victim) && !chan_master(victim) && !glob_bot(victim) && !match_my_nick(nick)) ameno_addkick(chan->name, nick);
	} else ameno_addkick(chan->name, nick);
    } 
}

/* a pain in the ass: mode changes */
static void gotmode(char *from, char *msg)
{
    char *nick, *ch, *op, *chg;
    char s[UHOSTLEN];
    char ms2[3];
    int z;
    struct userrec *u;
    memberlist *m;
    struct chanset_t *chan;

/* usermode changes? */
    if (msg[0] && (strchr(CHANMETA, msg[0]) != NULL)) {
        ch=newsplit(&msg);
	chg=newsplit(&msg);
	reversing=0;
	chan=findchan(ch);
	if (!chan) {
	    putlog(LOG_MISC, "*", CHAN_FORCEJOIN, ch);
	    dprintf(DP_SERVER, "PART %s\n", ch);
	    return;
	} else if (!channel_pending(chan)) {
	    z = strlen(msg);
	    if (msg[--z] == ' ') msg[z] = 0;
	    putlog(LOG_MODES, chan->name, "%s: mode change '%s %s' by %s", ch, chg, msg, from);
	    u=get_user_by_host(from);
	    get_user_flagrec(u, &user, ch);
	    nick=splitnick(&from);
	    m=ismember(chan, nick);
	    if (m) m->last=now;
	} else return;
    } else return;
    ms2[0]='+';
    ms2[2]=0;
    while ((ms2[1] = *chg)) {
	int todo = 0;
	switch (*chg) {
	case '+':	ms2[0] = '+'; break;
	case '-':	ms2[0] = '-'; break;
	case 'i':	todo=CHANINV;
			if (!nick[0]) reversing=1;
			break;
	case 'p':	todo=CHANPRIV;
			if (!nick[0]) reversing=1;
			break;
	case 's':	todo=CHANSEC;
			if (!nick[0]) reversing=1;
			break;
	case 'm':	todo=CHANMODER;
			if (!nick[0]) reversing=1;
			break;
	case 't':	todo = CHANTOPIC;
			if (!nick[0]) reversing=1;
			break;
	case 'n':	todo = CHANNOMSG;
			if (!nick[0]) reversing=1;
			break;
	case 'a':	todo = CHANANON;
			if (!nick[0]) reversing=1;
			break;
	case 'q':	todo = CHANQUIET;
			if (!nick[0]) reversing=1;
			break;
	case 'l':	if (!nick[0]) reversing = 1;
			if (ms2[0] == '-') {
			    check_tcl_mode(nick, from, u, chan->name, ms2, "");
			    if ((reversing) && (chan->channel.maxmembers != (-1)) && ((rand() % 100) < ameno_op_prc)) {
				simple_sprintf(s, "%d", chan->channel.maxmembers);
				add_mode(chan, '+', 'l', s);
			    }
			    chan->channel.maxmembers=(-1);
			} else {
			    op=newsplit(&msg);
			    fixcolon(op);
			    if (op == '\0') break;
			    chan->channel.maxmembers=atoi(op);
			    check_tcl_mode(nick, from, u, chan->name, ms2, int_to_base10(chan->channel.maxmembers));
			    if (reversing && ((rand() % 100) < ameno_op_prc)) add_mode(chan, '-', 'l', "");
			}
			break;
	case 'k':	if (!nick[0]) reversing=1;
			if (ms2[0] == '+') chan->channel.mode|=CHANKEY; else chan->channel.mode&=~CHANKEY;
			op=newsplit(&msg);
			fixcolon(op);
			if (op == '\0') break;
			if (ms2[0] == '+') got_key(chan, nick, from, op);
			else {
			    if ((reversing) && (chan->channel.key[0])) {
			      if ((rand() % 100) < ameno_op_prc) add_mode(chan, '+', 'k', chan->channel.key);
			    } else if ((chan->key_prot[0]) && !glob_master(user) && !chan_master(user) && ((rand() % 100) < ameno_op_prc)) {
			      if ((rand() % 100) < ameno_op_prc) add_mode(chan, '+', 'k', chan->key_prot);
			    }
			    set_key(chan, NULL);
			}
			break;
	case 'o':	op=newsplit(&msg);
			fixcolon(op);
			if (ms2[0] == '+') got_op(chan, nick, from, op, u, &user); 
			else got_deop(chan, nick, from, op, u);
			break;
	case 'v':	break;
	case 'b':	op=newsplit(&msg);
			fixcolon(op);
			check_tcl_mode(nick, from, u, chan->name, ms2, op);
			if (ms2[0] == '+') got_ban(chan, nick, from, op);
			else got_unban(chan, nick, from, op, u);
			break;
	case 'e':	op=newsplit(&msg);
			fixcolon(op);
			check_tcl_mode(nick, from, u, chan->name, ms2, op);
			if (ms2[0] == '+') got_exempt(chan, nick, from, op);
			else got_unexempt(chan, nick, from, op, u);
			break;
	case 'I':	op = newsplit(&msg);
			fixcolon(op);
			check_tcl_mode(nick, from, u, chan->name, ms2, op);
			if (ms2[0] == '+') got_invite(chan, nick, from, op);
			else got_uninvite(chan, nick, from, op, u);
			break;
	}
	if (todo) {
	    check_tcl_mode(nick, from, u, chan->name, ms2, "");
	    if (ms2[0] == '+') chan->channel.mode |= todo; else chan->channel.mode &= ~todo;
	    if ((((ms2[0] == '+') && (chan->mode_mns_prot & todo)) || ((ms2[0] == '-') && (chan->mode_pls_prot & todo))) && !glob_master(user) && !chan_master(user)) {
		if ((rand() % 100) < ameno_op_prc) add_mode(chan, ms2[0] == '+' ? '-' : '+', *chg, "");
	    } else if (reversing && ((ms2[0] == '+') || (chan->mode_pls_prot & todo)) && ((ms2[0] == '-') || (chan->mode_mns_prot & todo))) {
		if ((rand() % 100) < ameno_op_prc) add_mode(chan, ms2[0] == '+' ? '-' : '+', *chg, "");
	    }
	}
	chg++;
    }
    if (!me_op(chan) && !nick[0]) chan->status|=CHAN_ASKEDMODES;
    ameno_flushrand(ameno_kick_prc / 3);
    ameno_flushrandop(ameno_op_prc / 3);
}
